/***********************************************************************
 *  Copyright (c) 2007 Novell, Inc.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ***********************************************************************/

/***********************************************************************/
//  Standard Boilerplate
//
const I = Components.interfaces;

function getService(ID, Interface) { 
	return Components.classes[ID].getService(Components.interfaces[Interface]);
}

function createInstance(ID, Interface) {
	return Components.classes[ID].createInstance(Components.interfaces[Interface]);
}

function log() {
	const Console = getService('@mozilla.org/consoleservice;1', 'nsIConsoleService');
	for (var args = ['MhtmlUnarchiver'], i = 0; i < arguments.length; args.push(arguments[i++]));
	Console.logStringMessage(args.join(' '));
}

/***********************************************************************/
//
// MhtmlProtocolHandler - static object responsible for handling the 
// x-mhtml-resource: protocol, so that resources associated with mhtml
// archives can be loaded with their proper content types.  Implementing
// a protocol handler is also necessary because Mozilla does not allow
// content loaded from the web to reference file: urls.  This object is
// also responsible for mapping text/plain 

var MhtmlProtocolHandler = new Object();

// XPCOM Bits
MhtmlProtocolHandler.NAME = 'MHTML Resource Protocol Handler';
MhtmlProtocolHandler.COMP_ID = Components.ID('{14013F0F-CDFF-4DC0-AC15-314AD2235CB4}');
MhtmlProtocolHandler.CONTRACT_ID = '@mozilla.org/network/protocol;1?name=x-mhtml-resource';
MhtmlProtocolHandler.wrappedJSObject = MhtmlProtocolHandler;

// nsIProtocolHandler implementation
MhtmlProtocolHandler.scheme = 'x-mhtml-resource';
MhtmlProtocolHandler.defaultPort = -1;
MhtmlProtocolHandler.protocolFlags = I.nsIProtocolHandler.URI_NORELATIVE | I.nsIProtocolHandler.URI_NOAUTH | I.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE;
MhtmlProtocolHandler._resourceMap = new Object();
MhtmlProtocolHandler._resourceList = new Array();

MhtmlProtocolHandler.allowPort = function(port, scheme) { return false };

MhtmlProtocolHandler.newURI = function(spec, charset, baseURI) {
	var uri = createInstance('@mozilla.org/network/simple-uri;1', 'nsIURI');
	uri.spec = spec;
	return uri;
};

MhtmlProtocolHandler.newChannel = function(uri) {
	var parts = uri.spec.match(/^x-mhtml-resource:(\d+)\/(\d+)/);
	var record = (parts) ? this._resourceList[Number(parts[2])] : null;
	
	// the token serves as a shared secret, so resource uris are not
	// guessable and thus retrievable outside the archive they came from.
	if(!parts || !record || parts[1] != record.token) return null;

	var ios = getService('@mozilla.org/network/io-service;1', 'nsIIOService');
	var file = createInstance('@mozilla.org/file/local;1', 'nsILocalFile');
	file.initWithPath(record.path);
	var channel = ios.newChannelFromURI(ios.newFileURI(file));
	if (record.contentType) 
		channel.contentType = record.contentType;
	return channel;
};

MhtmlProtocolHandler.store = function(token, uri, contentType, path) {
	var record = { 
		contentType: contentType, 
		path: path,
		token: token,
		index: this._resourceList.length
	};
	this._resourceMap[token + ':' + this._normalizeURI(uri)] = record;
	this._resourceList.push(record);
};

MhtmlProtocolHandler.retrieve = function(token, uri) {
	var hash = '', index;
	if ((index = uri.indexOf('#')) != -1) {
		hash = uri.slice(index);
		uri = uri.slice(0, index);
	}
	var record = this._resourceMap[token + ':' + this._normalizeURI(uri)];
	return (record) ? this.scheme + ':' + token + '/' + record.index + hash : uri + hash;
};

MhtmlProtocolHandler._normalizeURI = function(uri) {
	return uri.replace(/&amp;/g, '&'); // todo, add more normalizations?
};

MhtmlProtocolHandler.rewriteHTMLUrls = function(unprocessedString, baseUrl, token) {
	var m, u, result = new Array();
	var tagexp = new RegExp('<[\s]*[a-z]+[^><]*>', 'gi');
	var urlattribre = new RegExp("(action|background|cite|classid|codebase|data|href|longdesc|profile|src|usemap){1}"
								+ "[\s]*=[\s]*(('[^'<>]+')|(\"[^\"<>]+\")|([^'\"<>\s]+)){1}", 'i');
	for (var index = 0; m = tagexp.exec(unprocessedString); index = tagexp.lastIndex) {
		result.push(unprocessedString.substring(index, m.index));
		var tag = tagstr = m.toString();
	
		// Now we have the tag, find the attribute in it that has a url
		var newTagstr = new Array();
		while (u = urlattribre.exec(tagstr)) {
			newTagstr.push(tagstr.substring(0, u.index));
			var urlattribstr = u[0].toString();
			var newUrlattribstr = urlattribstr.substring(0, urlattribstr.indexOf('=') + 1);
			newTagstr.push(newUrlattribstr);
			var attribstr = trim(urlattribstr.slice(urlattribstr.indexOf('=') + 1));

			if (attribstr[0] == '"' || attribstr[0] == "'") {
				var quote = attribstr[0];
				newTagstr.push(quote, this.retrieve(token, baseUrl.resolve(attribstr.slice(1, -1))), quote);
			} else {
				newTagstr.push(this.retrieve(token, baseUrl.resolve(attribstr)));
			}
			
			tagstr = tagstr.slice(u.index + urlattribstr.length);
		}
	
		newTagstr.push(tagstr);
		newTagstr = newTagstr.join('');
		result.push(newTagstr);
		// add a base tag to the head so that relative resources may be found;
		// This shouldn't conflict with an existing base because IE appears to 
		// absolutize URLs and remove the <base> when writing archives.
		if (newTagstr.match(/<head/i)) {
			result.push('<base href="' + baseUrl.spec + '">');
		}

		// if the tag is a style, rewrite imports
		if (tag.match(/<style/im)){
			var start = tagexp.lastIndex, end = new RegExp('<\\/style', 'gim');
			end.lastIndex = start;
			if (end = end.exec(unprocessedString)) {
				var styles = unprocessedString.slice(start, end.index);
				result.push(this.rewriteCSSUrls(styles, baseUrl, token));
				tagexp.lastIndex = end.index;
			}
		}
	}
	result.push(unprocessedString.slice(index));
	return result.join('');
};

MhtmlProtocolHandler.rewriteCSSUrls = function(data, base, token) {
	var rewrite = function(match, url) {
		return match.replace(url, MhtmlProtocolHandler.retrieve(token, base.resolve(url)));
	};
	return data.replace(/url\(\s*(\S*)\s*\)/gim, rewrite);
};

MhtmlProtocolHandler.makeTempDirectory = function(hint) {
	try {
		if (!this.tmp) {
			// System-provided temp directory
			var tmp = getService('@mozilla.org/file/directory_service;1', 'nsIProperties')
				.get('TmpD', I.nsIFile);
	
			// Per-process temp directory - contents of this directory will 
			// currently "leak" if Firefox doesn't shut down cleanly. Multiple
			// Firefoxes running under the same (or different) user will not
			// conflict however.
			tmp.append('com.novell.mhtml');
			tmp.createUnique(I.nsIFile.DIRECTORY_TYPE,  0700);
			this.tmp = tmp;
			getService('@mozilla.org/observer-service;1', 'nsIObserverService')
				.addObserver(this, 'quit-application', false);
		}
		var name = 'archive';
		if (hint) {
			var names = hint.split('/');
			while (names.length && !names[names.length-1]) names.pop();
			name = names.pop() || name;
		}	
		tmp = this.tmp.clone();
		tmp.append(name)
		tmp.createUnique(I.nsIFile.DIRECTORY_TYPE,  0700);
		return tmp;
	} catch (e) {
		log('Error: failed to create temporary directory', tmp, 'with error', e);
		throw e;
	}
};

// nsIObserver implementation
MhtmlProtocolHandler.observe = function(subject, topic, data) {
	// Listen for http responses
	if (topic == 'app-startup') {
		getService('@mozilla.org/observer-service;1', 'nsIObserverService')
			.addObserver(this, 'http-on-examine-response', false);

	// Remap mime-type of *.mht and *.mhtml served with text/plain 
	} else if (topic == 'http-on-examine-response') {
		const mime_match = new RegExp('^text\\/plain');
		const url_match = new RegExp('\\.mht$|\\.mhtml$', 'i');
		var channel = subject.QueryInterface(I.nsIHttpChannel);
		if (mime_match(channel.contentType) && url_match(channel.URI.spec)) {
			channel.contentType = 'message/rfc822';
			// If we let Mozilla cache an archive for which we've overwritten
			// the header, it will not inform us the next time it tries to load
			// it for some reason --- perhaps it's not caching the changed 
			// header?
			channel.setResponseHeader('Cache-Control', 'no-store', false);
		}

	// recursively delete the temp directory on quit	
	} else if (topic == 'quit-application' && this.tmp) {
		this.tmp.remove(true);
	}
};

// nsISupports implementation (skipping addRef and Release methods)
MhtmlProtocolHandler.QueryInterface = function(iid) {
	if (!iid.equals(I.nsISupports) && !iid.equals(I.nsIProtocolHandler)
		&& !iid.equals(I.nsIObserver)){
			throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
};

/***********************************************************************/
//
//  Utility Functions
//

function trim(str) {
	return str.replace(/^\s*(.*)/, '$1').replace(/(.*?)\s*$/, '$1');
}

function newURI(spec) {
	var uri = createInstance('@mozilla.org/network/standard-url;1', 'nsIURI');
	try {
		uri.spec = spec;
	} catch(e) {
		// IE has no compunctions about putting unparsable URLs in mhtml (e.g. file urls)
		var urlparts = spec.split(':');
		uri.spec = urlparts[0] + ':' + encodeURIComponent(urlparts.slice(1));
		log('Warning: rewriting unparsable URI', spec, 'as', uri.spec);
	}
	return uri;
}

/***********************************************************************/
//
// XPCOM Boilerplate
//
var XPCOM = Object();

// nsISupports implementation (skipping addRef and Release methods)
XPCOM.QueryInterface = function(iid) {
	if (iid.equals(I.nsISupports) || iid.equals(I.nsIFactory) || iid.equals(I.nsIModule)) {
		return this;
	}
	throw Components.results.NS_ERROR_NO_INTERFACE;
};

// nsIFactory implementation (skipping lockFactory)
XPCOM.createInstance = function (outer, iid) {
	if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
	return MhtmlProtocolHandler.QueryInterface(iid);
};

// nsIModule implementation
XPCOM.registerSelf = function(compMgr, fileSpec, location, type) {
	compMgr.QueryInterface(I.nsIComponentRegistrar)
		.registerFactoryLocation(MhtmlProtocolHandler.COMP_ID, MhtmlProtocolHandler.NAME, MhtmlProtocolHandler.CONTRACT_ID, fileSpec, location, type);
	getService('@mozilla.org/categorymanager;1', 'nsICategoryManager')
		.addCategoryEntry('app-startup', MhtmlProtocolHandler.NAME, 'service,' + MhtmlProtocolHandler.CONTRACT_ID, true, true);
};

XPCOM.unregisterSelf = function(compMgr, location, type) {
	compMgr.QueryInterface(I.nsIComponentRegistrar)
		.unregisterFactoryLocation(MhtmlProtocolHandler.COMP_ID, location);
	getService('@mozilla.org/categorymanager;1', 'nsICategoryManager')
		.deleteCategoryEntry('app-startup', MhtmlProtocolHandler.NAME, true);
};

XPCOM.getClassObject = function(compMgr, comp_id, iid) {
	if (!iid.equals(I.nsIFactory)) throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	if (!comp_id.equals(MhtmlProtocolHandler.COMP_ID)) throw Components.results.NS_ERROR_NO_INTERFACE;
	return this.QueryInterface(I.nsIFactory);
};

XPCOM.canUnload = function(compMgr) { return true; };

// This is only called at component registration (first launch after install) 
function NSGetModule(compMgr, fileSpec) { return XPCOM; }