/***********************************************************************
 *  Copyright (c) 2007 Novell, Inc.
 *  Portions Copyright (c) 2005 Christopher Ottley (see http://maf.mozdev.org/)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ***********************************************************************/

/***********************************************************************/
//  Standard Boilerplate
//
const I = Components.interfaces;

function getService(ID, Interface) { 
	return Components.classes[ID].getService(Components.interfaces[Interface]);
}

function getJSService(ID) { 
	return Components.classes[ID].getService(Components.interfaces.nsISupports).wrappedJSObject;
}

function createInstance(ID, Interface) {
	return Components.classes[ID].createInstance(Components.interfaces[Interface]);
}

function log() {
	const Console = getService('@mozilla.org/consoleservice;1', 'nsIConsoleService');
	for (var args = ['MhtmlUnarchiver'], i = 0; i < arguments.length; args.push(arguments[i++]));
	Console.logStringMessage(args.join(' '));
}

/***********************************************************************/
//
//  MhtmlUnarchiver - implements an nsIStreamConverter (and consequently
//  an nsIStreamConverter and nsIRequestObserver) which converts incoming
//  message/rfc822 data into a text/html output stream that can be
//  displayed by the browser.

function MhtmlUnarchiver() { this.wrappedJSObject = this; }

// XPCOM bits
MhtmlUnarchiver.NAME = 'MHTML Unarchiver Component';
MhtmlUnarchiver.COMP_ID = Components.ID('{3EAF967C-0276-4822-8448-5CCEB56A4DD4}');
MhtmlUnarchiver.CONTRACT_ID1 = '@mozilla.org/streamconv;1?from=message/rfc822&to=*/*';
MhtmlUnarchiver.CONTRACT_ID2 = '@mozilla.org/streamconv;1?from=application/x-mhtml-archive&to=*/*';

MhtmlUnarchiver.prototype = new Object();

// nsISupports implementation (skipping addRef and Release methods)
MhtmlUnarchiver.prototype.QueryInterface = function(iid) {
	if (!iid.equals(I.nsISupports) && !iid.equals(I.nsIStreamConverter) &&
		!iid.equals(I.nsIStreamListener) && !iid.equals(I.nsIRequestObserver)){
			throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
};

// nsIStreamConverter implementation
MhtmlUnarchiver.prototype.asyncConvertData = function(fromType, toType, listener, context) {
	this.listener = listener;
};

MhtmlUnarchiver.prototype.convert = function(fromStream, fromType, toType, context) {
	// don't bother implementing synchronous conversion
	throw Components.results.NS_ERROR_NOT_IMPLEMENTED; 
};

// nsIRequestObserver && nsIStreamListener implementation
MhtmlUnarchiver.prototype.onStartRequest = function(request, context) {
	this.buffer = new Array();
	this.instream = createInstance('@mozilla.org/scriptableinputstream;1', 'nsIScriptableInputStream');
	request.QueryInterface(I.nsIChannel).contentType = 'text/html';
	this.listener.QueryInterface(I.nsIRequestObserver).onStartRequest(request, context);
};

MhtmlUnarchiver.prototype.onDataAvailable = function(request, context, instream, offset, count) {
	if (offset == 0) this.instream.init(instream);
	this.buffer.push(this.instream.read(count));
};

MhtmlUnarchiver.prototype.onStopRequest = function(request, context, statusCode) {
	this.instream.close();
	delete this.instream;

	if (statusCode) { // bail
		delete this.buffer;
		return; 
	}

	// extract archive
	this.buffer = this.buffer.join('');
	this.buffer = this.extractArchive(this.buffer, request.name);

	// push the extracted archive out to our listener
	var strstream = createInstance('@mozilla.org/io/string-input-stream;1', 'nsIStringInputStream');
	strstream.setData(this.buffer, -1);
	this.listener.QueryInterface(I.nsIStreamListener)
		.onDataAvailable(request, context, strstream, 0, this.buffer.length);
	this.listener.QueryInterface(I.nsIRequestObserver)
		.onStopRequest(request, context, statusCode);
};

MhtmlUnarchiver.prototype.extractArchive = function(content, hint) {
	const protocolHandler = getJSService('@mozilla.org/network/protocol;1?name=x-mhtml-resource');
	var destpath = protocolHandler.makeTempDirectory(hint);
	var state = {
		htmlFiles: new Array(),
		htmlBases: new Array(),
		cssFiles: new Array(),
		cssBases: new Array(),
		token: String(Math.floor(Math.random() * Math.pow(2, 64)))
	};

	// decode parts
	new MhtmlDecoder(content).extractContent(destpath, state);

	// Rewrite URLs
	for (var i = 0; i < state.htmlFiles.length; i++) {
		var thisPage = readFile(state.htmlFiles[i]);
		thisPage = protocolHandler.rewriteHTMLUrls(thisPage, newURI(state.htmlBases[i]), state.token);
		state.htmlFiles[i].remove(false);
		if (i==0) {
			var result = thisPage;
		} else {
			createUniqueFile(state.htmlFiles[i], thisPage);
		}
	}
	for (var i = 0; i < state.cssFiles.length; i++) {
		var srcSheet = readFile(state.cssFiles[i]);
		var dstSheet = protocolHandler.rewriteCSSUrls(srcSheet, newURI(state.cssBases[i]), state.token);
		if (srcSheet !== dstSheet) {
			state.cssFiles[i].remove(false);
			createUniqueFile(state.cssFiles[i], dstSheet);
		}
	}	
	return result;
};

/***********************************************************************/
//
// MhtmlDecoder
//

function MhtmlDecoder(content) {
	var unparsedBody = '', unparsedHeaders = content, splitIndex = -1;
	if ((splitIndex = content.indexOf('\r\n\r\n')) != -1) {
		unparsedHeaders = content.substring(0, splitIndex);
		unparsedBody = content.slice(splitIndex + 4);
	} else if ((splitIndex = content.indexOf('\n\n')) != -1) {
		unparsedHeaders = content.substring(0, splitIndex);
		unparsedBody = content.slice(splitIndex + 2);
	}
	this.parseHeaders(unparsedHeaders);
	this.parseBody(unparsedBody);
};

MhtmlDecoder.prototype = new Object();

MhtmlDecoder.prototype.parseHeaders = function(unparsedHeaders) {
    // Ensure that values that cross lines end up on only one line
    var normalizedLines = this._normalizeHeaders(unparsedHeaders.split(/\r?\n/));
    this.headers = new Object();

    for (var i=0; i<normalizedLines.length; i++) {
      var delimiter = normalizedLines[i].indexOf(':');
      if (delimiter) {
      	var name = trim(normalizedLines[i].substring(0, delimiter).toLowerCase());
      	var value = trim(normalizedLines[i].slice(delimiter + 1));
      	this.headers[name] = value;
      }
    }
};

MhtmlDecoder.prototype._normalizeHeaders = function(headers) {
	// Put headers on a single line with content removing comments
	var result = new Array();
	var currentHeader = '', currentContent = '';
	var newHeader = '', newContent = '';

	for (var i=0; i<headers.length; i++) {
		var header = headers[i];
		var newHeader = trim(header.substring(0, header.indexOf(':')));
		var newContent = (header.indexOf(':') > 0)
			? trim(header.slice(header.indexOf(':') + 1))
			: trim(header);

		if (!currentHeader && !currentContent) {
			// First header
			currentHeader = newHeader;
			currentContent = this._removeCommentFromHeader(newHeader, newContent);
		} else if (currentHeader && !currentContent) {
			// Header with empty content
			currentContent = this._removeCommentFromHeader(currentHeader, trim(headers[i]));
		} else if (currentHeader && !newHeader) {
			// Content continues
			currentContent += this._removeCommentFromHeader(currentHeader, newContent);
		} else if (currentHeader && newHeader) {
			// We have a new current
			result[result.length] = currentHeader + ': ' + currentContent;
			currentHeader = newHeader;
			currentContent = this._removeCommentFromHeader(newHeader, newContent);
			newHeader = newContent = '';
		}		
	}

	if (currentHeader) {
		result[result.length] = currentHeader + ': ' + currentContent;
	}

	return result;
};

MhtmlDecoder.prototype._removeCommentFromHeader = function(name, value) {
	return (trim(name.toLowerCase()) != 'content-type') && (value.indexOf(';') > -1)
		? value.substring(0, value.indexOf(';'))
		: value;
};

MhtmlDecoder.prototype.parseBody = function(unparsedBody) {
	this.body = new Array(), this.parts = new Array();
	// If the content type isn't multipart
	if ((this.getHeaderValue('content-type').indexOf('multipart')) == -1) {
		this.body.push(unparsedBody);
		this.parts.push(this);
		return;
	}

	// Get the boundary string and split the parts up
	var ctData = this._convertContentType(this.getHeaderValue('content-type'));
	var boundary = start = '';
	for (var i=0; i<ctData.length; i++) {
		var entry = ctData[i];
		if (entry.name == 'boundary') {
			boundary = entry.value;
		} else if (entry.name == 'start') {
			start = entry.value;
		}
	}

	// If we have a boundary, split
	if (boundary) {
		// Remove the first boundary and the last boundary marker
		var contentBodyLessBoundaryEnd = unparsedBody;
		var boundaryEnd = contentBodyLessBoundaryEnd.indexOf('--' + boundary + '--');
		var boundaryStart = contentBodyLessBoundaryEnd.indexOf('--' + boundary) + boundary.length + 2;
		if (boundaryEnd > 0) {
			contentBodyLessBoundaryEnd = contentBodyLessBoundaryEnd.substring(boundaryStart, boundaryEnd);
		}

		var contentParts = contentBodyLessBoundaryEnd.split('--' + boundary);
		for (var i=0; i<contentParts.length; i++) {
			var currentPart = contentParts[i];
			this.body.push(currentPart);
			this.parts.push(new MhtmlDecoder(currentPart)); 
		}
	}
};

MhtmlDecoder.prototype._convertContentType = function(contentType) {
	var result = new Array();
	var ctArray = contentType.split(';');

	// Get the name value pairs
	for (var i=0; i<ctArray.length; i++) {
		var entry = trim(ctArray[i]), delim = entry.indexOf('=');
		if (delim > -1) {
			result.push({
				name: trim(entry.substring(0, delim).toLowerCase()),
				value: entry.substring(delim + 2, entry.length-1) // remove quotes
			});
		}
	}

	return result;
};

MhtmlDecoder.prototype.getHeaderValue = function(headerName) {
	return this.headers[trim(headerName.toLowerCase())] || '';
};

MhtmlDecoder.prototype._findRootPartNo = function() {
	var ctData = this._convertContentType(this.getHeaderValue('content-type'));	
	var needleType;
	for (var i=0; i<ctData.length; i++) {
		var entry = ctData[i];
		if (entry.name == 'type') {
			needleType = trim(entry.value.toLowerCase());
		}
	}
	if (!needleType) return 0;

	for (var i=0; i<this.body.length; i++) {
		var types = this.parts[i].getHeaderValue('content-type').split(';');
		for (var j=0; j<types.length; j++) {
			if (types[j] == needleType) {
				return i;
			}
		}
	}
};

MhtmlDecoder.prototype.extractContent = function(destpath, state) {
	if (this.parts.length > 1) {
		// extracting the root first ensures state.htmlFiles[0] is the root
		var rootPartNo = this._findRootPartNo();
		this.parts[rootPartNo].extractContent(destpath, state);
		for(var i=0; i<this.parts.length; i++) {
			if (i != rootPartNo)
				this.parts[i].extractContent(destpath, state);
		}
		return;
	}
	
	var contentLocation = this.getHeaderValue('Content-Location');
	var contentType = this.getHeaderValue('Content-Type');
	var contentId = this.getHeaderValue('Content-Id');
	if (!contentLocation && !contentId) return;
	
	// Determine filename
	var contentURI = newURI(contentLocation);
	filename = contentURI.path;
	if (filename.indexOf('?') != -1)
		filename = filename.slice(0, filename.indexOf('?'));
	if (filename[filename.length - 1] == '/')
		filename = filename.slice(0, -1);
	filename = filename.slice(filename.lastIndexOf('/') + 1);
	// Prevent linux from thinking it knows the type of these files
	if (contentType == 'application/octet-stream' && /(\.aspx?$|\.php$)/i.test(filename))
		filename = filename.replace(/(\.aspx?$|\.php$)/i, '');
	var destfile = destpath.clone();
	destfile.append(filename);

	// Write the file
	var data = this.body[0], encoding = this.getHeaderValue('Content-Transfer-Encoding');
	if (encoding == 'quoted-printable') {
		data = this._decodeQuotedPrintable(data, contentType);
	} else if (encoding == 'base64') try {
		data = getService('@mozilla.org/appshell/appShellService;1', 'nsIAppShellService')
			.hiddenDOMWindow.atob(data.split(/\r?\n/).join(''));
	} catch (e) {
		data = '';
		log('Error: invalid binary content at location', contentLocation);
	}
	createUniqueFile(destfile, data);

	if (contentType == 'application/octet-stream') try {
		// IE frequently lies about the content-type it got from the server,
		// so we might as well sniff the extension to catch images/script/css;
		// sniffing the data would be better still, but I can't get any 
		// nsIContentSniffer to instantiate
		contentType = getService('@mozilla.org/mime;1', 'nsIMIMEService').getTypeFromURI(contentURI);
		contentType = /(aspx?$|php$)/.test(contentType) ? undefined :contentType; 
	} catch (e) {
		log('Warning: Setting undefined content-type for resource at location', contentLocation || contentId); 
		contentType = undefined;
	}
	
	if (contentType && startsWith(contentType, 'text/html')) {
		state.htmlFiles.push(destfile);
		state.htmlBases.push(contentLocation);
	} else if (contentType && startsWith(contentType, 'text/css')) {
		state.cssFiles.push(destfile);
		state.cssBases.push(contentLocation);
	}
	const protocolHandler = getJSService('@mozilla.org/network/protocol;1?name=x-mhtml-resource');
	if (contentLocation) {
		protocolHandler.store(state.token, contentLocation, contentType, destfile.path);
	}
	if (contentId) try {
		var cidURL = 'cid:' + encodeURIComponent(contentId.match(/<(.*?)>/)[1]);
		protocolHandler.store(state.token, cidURL, contentType, destfile.path);
	} catch (e) {
		log('Warning: encountered bogus content-id', contentId);
	}
};

MhtmlDecoder.prototype._decodeQuotedPrintable = function(content, contentType) {
	// Loose check to see if charset is UTF, should really check charset
	if (contentType && /utf-/i.test(contentType) && startsWith(content, '=EF=BB=BF')) {
		content = content.slice('=EF=BB=BF'.length);
	}
	
	// = sign followed by new line, replaced by nothing.
	content = content.replace(/=\r?\n/g, '');
	
	var lines = content.split('=');
	var result = [lines[0]];
	for (var i=1; i<lines.length; i++) {
		if (lines[i].length > 1) {
			result.push(String.fromCharCode(parseInt(lines[i].substring(0,2), 16)));
			if (lines[i].length > 2) {
				result.push(lines[i].slice(2));
			}
		}
	}
	return result.join('');
};

/***********************************************************************/
//
//  Utility Functions
//

function trim(str) {
	return str.replace(/^\s*(.*)/, '$1').replace(/(.*?)\s*$/, '$1');
}

function startsWith(str, substr) {
	return (str.substring(0, substr.length) == substr);
}

function createUniqueFile(file, contents) {
	try {
		file.createUnique(I.nsIFile.NORMAL_FILE_TYPE, 0600);
		var stream = createInstance('@mozilla.org/network/file-output-stream;1', 'nsIFileOutputStream');
		stream.init(file, -1, -1, 0); // write, create, truncate
		stream.QueryInterface(I.nsIOutputStream).write(contents, contents.length);
		stream.close();
	} catch (e) {
		log('Error: failed to create file', file.path, 'with error', e);
	}
}

function readFile(file) {
	try {
		var instream = createInstance('@mozilla.org/network/file-input-stream;1', 'nsIFileInputStream')
		instream.init(file, 0x01, 0444, null);
		
		var sinstream = createInstance('@mozilla.org/scriptableinputstream;1', 'nsIScriptableInputStream')
		sinstream.init(instream);
		var result = sinstream.read(file.fileSize);
		sinstream.close();
		instream.close();
		return result;
	} catch (e) {
		log('Error: failed to read file', file.path, 'with error', e);
		return '';
	}
}

function newURI(spec) {
	var uri = createInstance('@mozilla.org/network/standard-url;1', 'nsIURI');
	try {
		uri.spec = spec;
	} catch(e) {
		// IE has no compunctions about putting unparsable URLs in mhtml (e.g. file urls)
		var urlparts = spec.split(':');
		uri.spec = urlparts[0] + ':' + encodeURIComponent(urlparts.slice(1));
		log('Warning: rewriting unparsable URI', spec, 'as', uri.spec);
	}
	return uri;
}

/***********************************************************************/
//
// XPCOM Boilerplate
//
var XPCOM = Object();

// nsISupports implementation (skipping addRef and Release methods)
XPCOM.QueryInterface = function(iid) {
	if (iid.equals(I.nsISupports) || iid.equals(I.nsIFactory) || iid.equals(I.nsIModule)) {
		return this;
	}
	throw Components.results.NS_ERROR_NO_INTERFACE;
};

// nsIFactory implementation (skipping lockFactory)
XPCOM.createInstance = function (outer, iid) {
	if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
	return (new MhtmlUnarchiver()).QueryInterface(iid);
};

// nsIModule implementation
XPCOM.registerSelf = function(compMgr, fileSpec, location, type) {
	compMgr.QueryInterface(I.nsIComponentRegistrar)
		.registerFactoryLocation(MhtmlUnarchiver.COMP_ID, MhtmlUnarchiver.NAME, MhtmlUnarchiver.CONTRACT_ID1, fileSpec, location, type);
	compMgr.QueryInterface(I.nsIComponentRegistrar)
		.registerFactoryLocation(MhtmlUnarchiver.COMP_ID, MhtmlUnarchiver.NAME, MhtmlUnarchiver.CONTRACT_ID2, fileSpec, location, type);

	// Register MhtmlUnarchiver component as a handler for MHTML
	var cm = getService('@mozilla.org/categorymanager;1', 'nsICategoryManager');
	cm.addCategoryEntry('ext-to-type-mapping', 'mht', 'message/rfc822', true, true);
	cm.addCategoryEntry('ext-to-type-mapping', 'mhtml', 'message/rfc822', true, true);
};

XPCOM.unregisterSelf = function(compMgr, location, type) {
	compMgr.QueryInterface(I.nsIComponentRegistrar)
		.unregisterFactoryLocation(MhtmlUnarchiver.COMP_ID, location);

	// Unregister this component as a handler for MHTML
	var cm = getService('@mozilla.org/categorymanager;1', 'nsICategoryManager');
	cm.deleteCategoryEntry('ext-to-type-mapping', 'mht', true);
	cm.deleteCategoryEntry('ext-to-type-mapping', 'mhtml', true);
};

XPCOM.getClassObject = function(compMgr, comp_id, iid) {
	if (!iid.equals(I.nsIFactory)) throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
	if (!comp_id.equals(MhtmlUnarchiver.COMP_ID)) throw Components.results.NS_ERROR_NO_INTERFACE;
	return this.QueryInterface(I.nsIFactory);
};

XPCOM.canUnload = function(compMgr) { return true; };

// This is only called at component registration (first launch after install) 
function NSGetModule(compMgr, fileSpec) { return XPCOM; }